
/* 
 * SubsetView.m -- Implementation file for the SubsetView class 
 *
 * You may freely copy, distribute, and reuse the code in this example.
 * NeXT disclaims any warranty of any kind, expressed or implied, as to 
 * its fitness for any particular use, nor even that it has any use.
 *
 */

#import "SusetView.h"
#import "nxlcau5h.h"
#import "nxardn5h.h"
#import "nxrijn5h.h"
#import <math.h> 
#import <stdlib.h>
#import <string.h>
#include <stdio.h>
#include <time.h>


extern parneih    amat;
extern parneih    bmat;
extern parneih    cmat;
extern parneih    dmat;
extern parneih    emat;

extern NXColor colst[KK];

int    suno=0;			/* node number in subset diagram 	*/
int    sust=0;			/* designated state in subset diagram	*/
int    suin=NO;			/* has subset diagram been prepared?	*/
int    fsubsback=YES;		/* draw opaque background?		*/
int    leenk[KK][DN];		/* links in de Bruijn diagram, by state	*/
int    liink[KK][SDN];		/* links in subset diagram, by state 	*/
int    sunanc[KK][SDN];		/* # ancestors in subset dgm, by state	*/
int    sudown[SDN];		/* depth of subset diagram nodes	*/
int    suhigh[SDN];		/* height of subset diagram nodes	*/
double scntr[SDN][2];		/* subset diagram coordinates		*/

void   asctobin();
int    closeto(double x, double y);
void   edSubset(int l, int pc, int pa, int pl, int pv);
void   howdown(int *sud);
void   ingather(double clst[][2], int *arry, int k, int n);
int    level (int n);
void   outgather(double clst[][2], int *arry, int k);
int    putlevel(twoparn *zmat, twoparn *xmat, int n);
void   suanc();
void   upthere(int *sud);
void   wru(char *c);

@implementation SubsetView

/* -----------------------------  S U B S E T  ---------------------------- */

/* put up the principal window for the SubsetView. Since subset.nib is a     */
/* secondary .nib file, it is not loaded until there is a demand, which    */
/* would usually arise from the "Subset" option in the main menu.	   */

- loadSubset:sender {
  if (subsetWindow==nil) {
    if (![NXApp loadNibSection: "subset.nib" owner: self withNames: NO]) 
	{return nil;}
    suin=NO;
    }
  [subsetWindow makeKeyAndOrderFront: nil];
  return self;
  }

- windowDidBecomeKey:sender {
	[menuCopy setTarget: self];
	[menuCopy setAction: @selector(copySubsetView:)];
	[menuCopy setEnabled: YES];
	[menuPrint setTarget: self];
	[menuPrint setAction: @selector(printSubset:)];
	[menuPrint setEnabled: YES];
return self;
}

- windowDidResignKey:sender {
	[menuCopy setEnabled: NO];
	[menuPrint setEnabled: NO];
return self;
}

- prepareSubset:sender {
char   c[WNML];
int    i, j, k, s, p;

asctobin();
wru(c);
[ruleNumber setStringValue: c];

zmat(&amat);  djmat(&amat,0);
zmat(&bmat);  djmat(&bmat,1);
zmat(&cmat);  djmat(&cmat,2);
zmat(&dmat);  djmat(&dmat,3);
zmat(&emat);  djmat(&emat,4);

/* generate the de Bruijn diagram's out-linkages */
for (i=0; i<DN; i++) {
  for (s=0, p=1, j=0; j<DN; j++, p*=2) if (amat.two[i][j]==1) s|=p;
  leenk[0][i]=s;
  for (s=0, p=1, j=0; j<DN; j++, p*=2) if (bmat.two[i][j]==1) s|=p;
  leenk[1][i]=s;
  for (s=0, p=1, j=0; j<DN; j++, p*=2) if (cmat.two[i][j]==1) s|=p;
  leenk[2][i]=s;
  for (s=0, p=1, j=0; j<DN; j++, p*=2) if (dmat.two[i][j]==1) s|=p;
  leenk[3][i]=s;
  for (s=0, p=1, j=0; j<DN; j++, p*=2) if (emat.two[i][j]==1) s|=p;
  leenk[4][i]=s;
  }

// for (i=0; i<DN; i++) printf("%3d ",leenk[0][i]);
// printf("\n");
// for (i=0; i<DN; i++) printf("%3d ",leenk[1][i]);
// printf("\n");

/* generate the subset diagram's out-linkages */
for (k=0; k<KK; k++) for (i=0; i<SDN; i++) {
  for (p=i, s=0, j=0; j<DN; p/=2, j++) if (p%2==1) s|=leenk[k][j];
  liink[k][i]=s;
  }
  
  suanc();
  howdown(sudown);
  [suTriv1 setIntValue: sudown[0]];
  [[suTriv2 cellAt:0 :0] setIntValue: sudown[  1]];
  [[suTriv2 cellAt:0 :1] setIntValue: sudown[  2]];
  [[suTriv2 cellAt:0 :2] setIntValue: sudown[  4]];
  [[suTriv2 cellAt:0 :3] setIntValue: sudown[  8]];
  [[suTriv2 cellAt:0 :4] setIntValue: sudown[ 16]];
  upthere(suhigh);
printf("suhigh[%2d]= %3d\n",SDN-1,suhigh[SDN-1]);
  scntr[0][0]=100.0;
  scntr[0][1]=420.0;
  for (i=1; i<SDN-1; i++) {
    scntr[i][0]=(double)(random()%200);
    scntr[i][1]=450.0-(double)(30+80*level(i))+(double)(random()%12);
    }
  scntr[SDN-1][0]=100.0;
  scntr[SDN-1][1]= 30.0;

  scntr[ 1][0]= 40.0;
  scntr[ 2][0]= 70.0;
  scntr[ 4][0]=100.0;
  scntr[ 8][0]=130.0;
  scntr[16][0]=160.0;

  scntr[30][0]= 40.0;
  scntr[29][0]= 70.0;
  scntr[27][0]=100.0;
  scntr[23][0]=130.0;
  scntr[15][0]=160.0;

  suin=YES;
  [self drawSubset: self];
return self;
}

- drawSubset:sender {[self display]; return self;}

- printSubset:sender {[subsetWindow smartPrintPSCode:self]; return self;}

- copySubsetView:sender {
NXRect zbounds;
id     pBoard;
  fsubsback=NO;
  pBoard = [Pasteboard new];
  [pBoard declareTypes: &NXPostScriptPboardType num: 1 owner: self];
  [theSubsetView getBounds: &zbounds];
  [theSubsetView writePSCodeInside: &zbounds to: pBoard]; 
  fsubsback=YES;
return self;
}

// - (BOOL)performDragOperation:(id <NXDraggingInfo>)sender
// {
//     Pasteboard *pboard = [sender draggingPasteboard];
// 
//     if (IncludesType([pboard types], NXColorPboardType)) {
// 	BOOL retval = NO;
// 	NXColor color = NXReadColorFromPasteboard(pboard);
// 	NXPoint p = [sender draggingLocation];
// 	[self convertPoint:&p fromView:nil];
// 	retval = [self acceptsColor:&color atPoint:&p];
// 	[NXApp updateWindows];	// reflect color change in Inspector, et. al.
// 	return retval;
//     }
// 
//     return YES;
// }

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* subroutines which actually do the drawing - - - - - - - - - - - - - - - - */

/* draw linkages &c in the subset diagram */
/*  l = state/color of link		  */
/* pc = draw circle?			  */
/* pa = draw arrow?			  */
/* pl = draw line?			  */
void edSubset(l,pc,pa,pl,pv) int l, pc, pa, pl, pv; {
int    i, j;
char   s[10];
double x, y, ex, wy, eks, wye, dx, dy, r, t;
  r=10.3;
  for (i=0; i<SDN; i++) {
	if (!pv && 
	    i != SDN-1 &&
	    (sunanc[0][i] == 0) &&
	    (sunanc[1][i] == 0) &&
	    (sunanc[2][i] == 0) &&
	    (sunanc[3][i] == 0) &&
	    (sunanc[4][i] == 0)
	    ) continue;
	ex=scntr[i][0];
	wy=scntr[i][1];
	PSmoveto(ex,wy);
	j=liink[l][i];
	x=scntr[j][0];
	y=scntr[j][1];
	dx=x-ex;
	dy=y-wy;
	t=atan2(dy,dx);
	eks=x-r*cos(t);
	wye=y-r*sin(t);
	if (pa==YES) {
	  PSlineto(eks,wye);
	  PSlineto(eks-0.1*dx+0.05*dy,wye-0.1*dy-0.05*dx);
	  PSlineto(eks-0.1*dx-0.05*dy,wye-0.1*dy+0.05*dx);
	  PSlineto(eks,wye);
	  }
	if (pc==YES) {
	  PSmoveto(x+10.3,y);
	  PSarc(x, y, 10.3, 0.0, 360.0);
	  PSmoveto(ex+10.3,wy);
	  PSarc(ex, wy, 10.3, 0.0, 360.0);
	  }

   if (pl==YES) {
     ex=scntr[i][0];
     wy=scntr[i][1];
     PSmoveto(ex-10.0,wy+5.0);
     sprintf(s,"%2d",i);
     PSshow(s);
     }
   }
}

/* calculate subset level */
int level (n) int n; {
int i, s;
s=0;
i=n;
while (i!=0) {s+=i%2; i/=2;}
return s;
}

/* calculate number of in-links for subset diagram nodes */
/* and place the results in the array sunanc		 */
void suanc() {
int i, j;
for (j=0; j<KK; j++) for (i=0; i<SDN; i++) sunanc[j][i]=0;
for (j=0; j<KK; j++) for (i=0; i<SDN; i++) sunanc[j][liink[j][i]]++;
}

/* list shortest path lengths from maximum to given node */
void howdown(sud) int *sud; {
int i, j, k, p;
  for (i=0; i<SDN; i++) sud[i]=SDN;
  sud[SDN-1]=0;
  for (j=0; j<SDN; j++) { 
    for (p=0, i=0; i<SDN; i++) for (k=0; k<KK; k++) 
	if (sud[liink[k][i]]>sud[i]+1) {sud[liink[k][i]]=sud[i]+1; p++;}
    if (p==0) break;
    }
}

/* list shortest path lengths from given node to minimum */
void upthere(sud) int *sud; {
int i, j, k, p;
  for (i=0; i<SDN; i++) sud[i]=SDN;
  sud[0]=0;
  for (j=0; j<SDN; j++) { 
    for (p=0, i=0; i<SDN; i++) for (k=0; k<KK; k++) 
	if (sud[i]>sud[liink[k][i]]+1) {sud[i]=sud[liink[k][i]]+1; p++;}
    if (p==0) break;
    }
}

/* find index of nearest center to x, y (subset) */
int closeto(x,y) double x, y; {int i, k; double ex, wy, r, p;
k=0; p=500000.0;
for (i=0; i<SDN; i++) {
  ex=x-scntr[i][0];
  wy=y-scntr[i][1];
  r=ex*ex+wy*wy;
  if (r<p) {p=r; k=i;}
  }
return k;
}

/* --------------------  S U B S E T   V I E W  ---------------------------- */

- initFrame:(const NXRect *)frameRect
/*
 * Initializes the new LiulGuph object.  First, an initFrame: message is sent 
 * to super to initialize LiulGuph as a View.  Next, the LiulGuph sets its own 
 * state -- that it is opaque and that the origin of its coordinate system lies 
 * in the center of its area.  It then creates and initializes its associated 
 * objects, a Storage object, an NXCursor, and a Cell.  Finally, it loads into 
 * the Window Server some PostScript procedures that it will use in drawing 
 * itself. 
 */
{
	NXPoint	spot;
	
	[super initFrame:frameRect];
	[self setOpaque:YES];
	crossCursor = [NXCursor newFromImage:
		[NXImage newFromSection:"cross.tiff"]];
	spot.x = spot.y = 7.0;
	[crossCursor setHotSpot:&spot];

  PSsetgray(NX_BLACK);
  PSselectfont("Times-Roman",16.0);

  [self setFlipped: YES];

	return self;

}


- setDelegate:anObject
/*
 * Sets the SubsetView's delegate instance variable to the supplied object.
 */
{
	delegate = anObject;
	return self;
}


- drawSelf:(const NXRect *)rects :(int)rectCount
/*
 * Draws the SubsetView's background and axes.  If there are any points,
 * these are drawn too.
 *
 */
{
        char             c[WNML], timday[30];
	unsigned int	 i, j, k;
  struct tm  *timeptr;
  time_t  secsnow;
	
	if (rects == NULL) return self;

	PSsetgray(NX_WHITE);
	if (fsubsback) NXRectFill(&rects[0]);
	PSsetgray(NX_DKGRAY);

   time(&secsnow);
   timeptr = localtime(&secsnow);
   sprintf(timday,"%d-%d-19%02d\n",
	((timeptr->tm_mon) + 1), 
	timeptr->tm_mday,
        timeptr->tm_year);

/* for the subset diagram -- - - - - - - - - - - - - - - - - - - - - - - - - */

   [[Font newFont: "Courier" size: 10.0 matrix:NX_FLIPPEDMATRIX] set];
   [timeDate setStringValue: timday];

	if (suin) {
	  for (i=0; i<KK; i++) {
	    if ([[linkAge cellAt: i: 0] state]) {
		PSsetlinewidth(1.0);
		NXSetColor(colst[i]);
		edSubset(i,NO,YES,NO,[treeShow state]);
		PSstroke();
		}
	    }
		PSsetlinewidth(0.0);
		PSsetgray(NX_BLACK); 
		edSubset(0,YES,NO,YES,[treeShow state]);
		PSstroke();
	}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	return self;
}

- sizeTo:(NXCoord)width :(NXCoord)height
/*
 * Ensures that whenever the SubsetView is resized, the origin of its 
 * coordinate system is repositioned to the center of its area.
 */
{
//	  [self setFlipped: YES];
	[super sizeTo:width :height];
	return self;
}

- mouseDown:(NXEvent *) theEvent
/*
 * Responds to a message the system sends whenever the user presses the mouse
 * button when the cursor is over the SubsetView.  The SubsetView changes the 
 * cursor to a cross-hairs image and then starts asking for mouse-dragged or 
 * mouse-up events. As it receives mouse-dragged events, the SubsetView updates 
 * the readOut text Cell with the cursor's coordinates.  If the user releases 
 * the mouse button while the cursor is over the SubsetView, the SubsetView 
 * registers the point and then sends a message to its delegate notifying it of 
 * the new point.
 */
{
	int	looping = YES, oldMask;
	NXPoint	aPoint;

         double  ex, wy;
	 int     i, j, k;

	[crossCursor set];

	oldMask = [window addToEventMask:NX_MOUSEDRAGGEDMASK];

/* for the subset diagram - - - - - - - - - - - - - - - - - - - - - - - - - */


 				[window setDocEdited:YES];
		                aPoint = theEvent->location;
		                [self convertPoint:&aPoint fromView:nil];
                                ex=aPoint.x;
                                wy=aPoint.y;
				suno=closeto(ex,wy);
				[nodeNumber setIntValue: suno];

       if (theEvent->data.mouse.click==3) {} else 
       if (theEvent->data.mouse.click==2) {for (i=0; i<KK; i++) {
	if ([[linkAge cellAt: i: 0] state]) 
	  switch ([[moveWhich selectedCell] tag]) {
	    case 0: ingather(scntr,liink[i],suno,SDN); break;
	    case 1: ingather(scntr,liink[i],liink[i][suno],SDN); break;
	    case 2: outgather(scntr,liink[i],suno); break;
	    case 3: outgather(scntr,liink[i],liink[i][suno]); break;
	    case 4: ingather(scntr,liink[i],suno,SDN);
		outgather(scntr,liink[i],suno); break;
	    default: break;}
		}
   [self display];

		} else {
	[self lockFocus];
         PSsetinstance(YES);
	do {
 		PSnewinstance();
		aPoint = theEvent->location;
		[self convertPoint:&aPoint fromView:nil];
                ex=aPoint.x;
                wy=aPoint.y;
				PSsetgray(NX_BLACK);
				PSmoveto(ex+12.0,wy);
				PSarc(ex, wy, 12.0, 0.0, 360.0);
				PSstroke();
				scntr[suno][0]=ex;
				scntr[suno][1]=wy;
		[window flushWindow];
		if (theEvent->type == NX_MOUSEUP) {
	    	/* on mouse-up, inform delegate, and clean up state */
			[window flushWindow];
			looping = NO;
		}	
   	} while (looping && (theEvent=[NXApp getNextEvent:
		NX_MOUSEUPMASK|NX_MOUSEDRAGGEDMASK]));
	[self unlockFocus];


   [self display];

	}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */ 

	[window setEventMask:oldMask];
	[NXArrow set];
	return self;
}

@end

/* end SubsetView.m */
